@@ -13,9 +13,12 @@ use anyhow::{Context, Result};
 use chrono::prelude::*;
 use prometheus::{IntCounter, IntGauge};
 use serde::{Deserialize, Deserializer};
-use std::convert::TryInto;
+use std::cell::Cell;
+use std::collections::BTreeSet;
 use std::fs;
+use std::rc::Rc;
 use std::time::Duration;
+use tokio::sync::RwLock;
 
 /// Default refresh interval for steady state (in seconds).
 pub(crate) const DEFAULT_STEADY_INTERVAL_SECS: u64 = 300; // 5 minutes.
@@ -92,7 +95,7 @@ where
 
 /// State machine for the agent.
 #[derive(Clone, Debug, PartialEq, Eq)]
-enum UpdateAgentState {
+enum UpdateAgentMachineState {
     /// Initial state upon actor start.
     StartState,
     /// Agent initialized.
@@ -121,15 +124,15 @@ enum UpdateAgentState {
     EndState,
 }
 
-impl Default for UpdateAgentState {
+impl Default for UpdateAgentMachineState {
     fn default() -> Self {
-        let start_state = UpdateAgentState::StartState;
+        let start_state = UpdateAgentMachineState::StartState;
         LATEST_STATE_CHANGE.set(chrono::Utc::now().timestamp());
         start_state
     }
 }
 
-impl UpdateAgentState {
+impl UpdateAgentMachineState {
     /// Progress the machine to a new state.
     fn transition_to(&mut self, state: Self) {
         use std::mem::discriminant;
@@ -142,10 +145,10 @@ impl UpdateAgentState {
 
     /// Transition to the Initialized state.
     fn initialized(&mut self) {
-        let target = UpdateAgentState::Initialized;
+        let target = UpdateAgentMachineState::Initialized;
         // Allowed starting states.
         assert!(
-            *self == UpdateAgentState::StartState,
+            *self == UpdateAgentMachineState::StartState,
             "transition not allowed: {:?} to {:?}",
             self,
             target,
@@ -156,10 +159,10 @@ impl UpdateAgentState {
 
     /// Transition to the ReportedSteady state.
     fn reported_steady(&mut self) {
-        let target = UpdateAgentState::ReportedSteady;
+        let target = UpdateAgentMachineState::ReportedSteady;
         // Allowed starting states.
         assert!(
-            *self == UpdateAgentState::Initialized,
+            *self == UpdateAgentMachineState::Initialized,
             "transition not allowed: {:?} to {:?}",
             self,
             target,
@@ -170,24 +173,26 @@ impl UpdateAgentState {
 
     /// Transition to the NoNewUpdate state.
     fn no_new_update(&mut self) {
-        let target = UpdateAgentState::NoNewUpdate;
+        let target = UpdateAgentMachineState::NoNewUpdate;
         // Allowed starting states.
         assert!(
-            *self == UpdateAgentState::ReportedSteady || *self == UpdateAgentState::NoNewUpdate,
+            *self == UpdateAgentMachineState::ReportedSteady
+                || *self == UpdateAgentMachineState::NoNewUpdate,
             "transition not allowed: {:?} to {:?}",
             self,
             target
         );
 
-        self.transition_to(UpdateAgentState::NoNewUpdate);
+        self.transition_to(UpdateAgentMachineState::NoNewUpdate);
     }
 
     /// Transition to the UpdateAvailable state with a new release.
     fn update_available(&mut self, update: Release) {
-        let target = UpdateAgentState::UpdateAvailable((update, 0));
+        let target = UpdateAgentMachineState::UpdateAvailable((update, 0));
         // Allowed starting states.
         assert!(
-            *self == UpdateAgentState::ReportedSteady || *self == UpdateAgentState::NoNewUpdate,
+            *self == UpdateAgentMachineState::ReportedSteady
+                || *self == UpdateAgentMachineState::NoNewUpdate,
             "transition not allowed: {:?} to {:?}",
             self,
             target
@@ -203,7 +208,7 @@ impl UpdateAgentState {
     /// (including the newly recorded failed attempt).
     fn record_failed_deploy(&mut self) -> (bool, u8) {
         let (release, attempts) = match self.clone() {
-            UpdateAgentState::UpdateAvailable((r, a)) => (r, a),
+            UpdateAgentMachineState::UpdateAvailable((r, a)) => (r, a),
             _ => unreachable!("transition not allowed: record_failed_deploy on {:?}", self,),
         };
         let fail_count = attempts.saturating_add(1);
@@ -220,14 +225,14 @@ impl UpdateAgentState {
 
     /// Transition to the UpdateAvailable state after a deploy failure.
     fn deploy_failed(&mut self, update: Release, fail_count: u8) {
-        let target = UpdateAgentState::UpdateAvailable((update, fail_count));
+        let target = UpdateAgentMachineState::UpdateAvailable((update, fail_count));
 
         self.transition_to(target);
     }
 
     /// Transition to the NoNewUpdate state after persistent deploy failure.
     fn update_abandoned(&mut self) {
-        let target = UpdateAgentState::NoNewUpdate;
+        let target = UpdateAgentMachineState::NoNewUpdate;
 
         self.transition_to(target);
     }
@@ -235,7 +240,7 @@ impl UpdateAgentState {
     /// Transition to the UpdateStaged state, setting the number of postponements
     /// remaining to `MAX_FINALIZE_POSTPONEMENTS`.
     fn update_staged(&mut self, update: Release) {
-        let target = UpdateAgentState::UpdateStaged((update, MAX_FINALIZE_POSTPONEMENTS));
+        let target = UpdateAgentMachineState::UpdateStaged((update, MAX_FINALIZE_POSTPONEMENTS));
 
         self.transition_to(target);
     }
@@ -243,18 +248,18 @@ impl UpdateAgentState {
     /// Determine whether to allow finalization based off of current state.
     /// Returns a boolean indicating whether a finalization is permitted.
     fn usersessions_can_finalize(&mut self) -> bool {
-        match get_interactive_user_sessions() {
-            Ok(interactive_sessions) => {
-                DETECTED_ACTIVE_USERS.set(interactive_sessions.len().try_into().unwrap());
-                self.handle_interactive_sessions(&interactive_sessions)
-            }
-            Err(e) => {
-                // If we failed to check for interactive sessions, just allow
-                // finalization.
-                log::error!("failed to check for interactive sessions: {}", e);
-                true
-            }
-        }
+        let interactive_sessions = get_interactive_user_sessions();
+
+        // If we failed to check for interactive sessions, assume nobody
+        // is logged in. Overall, we prefer forcibly finalizing updates
+        // instead of getting stuck in a retry loop here.
+        let sessions = interactive_sessions.unwrap_or_else(|e| {
+            log::error!("failed to check for interactive sessions: {}", e);
+            log::warn!("assuming no active sessions and proceeding anyway");
+            vec![]
+        });
+
+        self.handle_interactive_sessions(&sessions)
     }
 
     /// Helper for determining whether to allow a finalization by first checking whether
@@ -262,29 +267,44 @@ impl UpdateAgentState {
     /// state's remaining postponements (possibly broadcasting warning messages to active sessions).
     ///
     /// Returns a boolean indicating whether finalization is permitted.
-    fn handle_interactive_sessions(&mut self, interactive_sessions: &[InteractiveSession]) -> bool {
+    fn handle_interactive_sessions(&self, interactive_sessions: &[InteractiveSession]) -> bool {
+        DETECTED_ACTIVE_USERS.set(interactive_sessions.len() as i64);
+        log::trace!(
+            "handling interactive sessions, total: {}",
+            interactive_sessions.len()
+        );
+
+        // Happy path, nobody to wait for.
         if interactive_sessions.is_empty() {
+            log::debug!("no interactive sessions detected");
             return true;
         }
 
-        let (release, postponements_remaining) = match self.clone() {
-            UpdateAgentState::UpdateStaged((r, p)) => (r, p),
+        let (release, postponements_remaining) = match self {
+            UpdateAgentMachineState::UpdateStaged((r, p)) => (r, *p),
             _ => unreachable!(
                 "transition not allowed: handle_interactive_sessions on {:?}",
                 self,
             ),
         };
 
+        // Maximum grace period reached, not delaying any further.
         if postponements_remaining == 0 {
+            log::warn!("reached end of grace period while waiting for interactive sessions");
             return true;
         }
 
         if postponements_remaining == MAX_FINALIZE_POSTPONEMENTS {
             let max_reboot_delay_secs =
                 DEFAULT_POSTPONEMENT_TIME_SECS.saturating_mul(MAX_FINALIZE_POSTPONEMENTS as u64);
+            log::warn!(
+                "interactive sessions detected, entering grace period (maximum {})",
+                format_seconds(max_reboot_delay_secs)
+            );
             let warning_msg = format_reboot_warning(max_reboot_delay_secs, &release.version);
             broadcast(&warning_msg, interactive_sessions);
         } else if postponements_remaining == 1 {
+            log::warn!("last attempt to wait for the end of all interactive sessions");
             let warning_msg =
                 format_reboot_warning(DEFAULT_POSTPONEMENT_TIME_SECS, &release.version);
             broadcast(&warning_msg, interactive_sessions);
@@ -297,7 +317,7 @@ impl UpdateAgentState {
     /// postponements allowed by one) after a finalization postponement.
     fn record_postponement(&mut self) {
         let (release, postponements_remaining) = match self.clone() {
-            UpdateAgentState::UpdateStaged((r, p)) => (r, p),
+            UpdateAgentMachineState::UpdateStaged((r, p)) => (r, p),
             _ => unreachable!(
                 "transition not allowed: handle_interactive_sessions on {:?}",
                 self,
@@ -311,21 +331,21 @@ impl UpdateAgentState {
     /// Transition to the UpdateStaged state, setting the number of postponements
     /// remaining to postponements_remaining.
     fn reboot_postponed(&mut self, update: Release, postponements_remaining: u8) {
-        let target = UpdateAgentState::UpdateStaged((update, postponements_remaining));
+        let target = UpdateAgentMachineState::UpdateStaged((update, postponements_remaining));
 
         self.transition_to(target);
     }
 
     /// Transition to the UpdateFinalized state.
     fn update_finalized(&mut self, update: Release) {
-        let target = UpdateAgentState::UpdateFinalized(update);
+        let target = UpdateAgentMachineState::UpdateFinalized(update);
 
         self.transition_to(target);
     }
 
     /// Transition to the End state.
     fn end(&mut self) {
-        let target = UpdateAgentState::EndState;
+        let target = UpdateAgentMachineState::EndState;
 
         self.transition_to(target);
     }
@@ -334,10 +354,10 @@ impl UpdateAgentState {
     /// jitter should be added.
     fn get_refresh_delay(&self, steady_interval: Duration) -> (Duration, bool) {
         match self {
-            UpdateAgentState::ReportedSteady | UpdateAgentState::NoNewUpdate => {
+            UpdateAgentMachineState::ReportedSteady | UpdateAgentMachineState::NoNewUpdate => {
                 (steady_interval, true)
             }
-            UpdateAgentState::UpdateStaged((_, postponements)) => {
+            UpdateAgentMachineState::UpdateStaged((_, postponements)) => {
                 // If postponements is less than `MAX_FINALIZE_POSTPONEMENTS`, that means the current tick
                 // led to a postponment, and so we should add a delay of `DEFAULT_POSTPONEMENT_TIME_SECS`.
                 if *postponements < MAX_FINALIZE_POSTPONEMENTS {
@@ -346,7 +366,7 @@ impl UpdateAgentState {
                     (Duration::from_secs(DEFAULT_REFRESH_PERIOD_SECS), true)
                 }
             }
-            UpdateAgentState::EndState => (Duration::from_secs(END_INTERVAL_SECS), true),
+            UpdateAgentMachineState::EndState => (Duration::from_secs(END_INTERVAL_SECS), true),
             _ => (Duration::from_secs(DEFAULT_REFRESH_PERIOD_SECS), true),
         }
     }
@@ -355,6 +375,30 @@ impl UpdateAgentState {
 /// Update agent.
 #[derive(Debug)]
 pub(crate) struct UpdateAgent {
+    /// Current state of the agent.
+    /// We use an `Rc` because consumers of this field will likely need to
+    /// own it (e.g. consumers in futures).
+    state: Rc<RwLock<UpdateAgentState>>,
+    /// Timestamp of last state transition.
+    /// Behind `Rc` for same reason as above.
+    /// `Cell` is sufficient because the update_agent actor runs on a single thread (SystemArbiter).
+    state_changed: Rc<Cell<DateTime<Utc>>>,
+    /// Update agent's information.
+    info: UpdateAgentInfo,
+}
+
+/// Non-atomic read-write state of the update agent.
+#[derive(Debug, Default)]
+pub(crate) struct UpdateAgentState {
+    /// Current state of the agent state machine.
+    machine_state: UpdateAgentMachineState,
+    /// List of releases to ignore.
+    denylist: BTreeSet<Release>,
+}
+
+/// Read-only information about the update agent.
+#[derive(Debug, Clone)]
+pub(crate) struct UpdateAgentInfo {
     /// Whether to allow automatic downgrades.
     allow_downgrade: bool,
     /// Cincinnati service.
@@ -369,10 +413,6 @@ pub(crate) struct UpdateAgent {
     rpm_ostree_actor: Addr<RpmOstreeClient>,
     /// Update strategy.
     strategy: UpdateStrategy,
-    /// Current status for agent state machine.
-    state: UpdateAgentState,
-    /// Timestamp of last state transition.
-    state_changed: DateTime<Utc>,
 }
 
 impl UpdateAgent {
@@ -380,15 +420,17 @@ impl UpdateAgent {
     pub(crate) fn with_config(cfg: Settings, rpm_ostree_addr: Addr<RpmOstreeClient>) -> Self {
         let steady_secs = cfg.steady_interval_secs.get();
         Self {
-            allow_downgrade: cfg.allow_downgrade,
-            cincinnati: cfg.cincinnati,
-            enabled: cfg.enabled,
-            identity: cfg.identity,
-            rpm_ostree_actor: rpm_ostree_addr,
-            steady_interval: Duration::from_secs(steady_secs),
-            state: UpdateAgentState::default(),
-            strategy: cfg.strategy,
-            state_changed: chrono::Utc::now(),
+            state: Rc::new(RwLock::new(UpdateAgentState::default())),
+            state_changed: Rc::new(Cell::new(chrono::Utc::now())),
+            info: UpdateAgentInfo {
+                allow_downgrade: cfg.allow_downgrade,
+                cincinnati: cfg.cincinnati,
+                enabled: cfg.enabled,
+                identity: cfg.identity,
+                rpm_ostree_actor: rpm_ostree_addr,
+                steady_interval: Duration::from_secs(steady_secs),
+                strategy: cfg.strategy,
+            },
         }
     }
 }
@@ -478,9 +520,10 @@ fn format_reboot_warning(seconds: u64, release_ver: &str) -> String {
 
     format!(
         "New update {} is available and has been deployed.\n\
-        If permitted by the update strategy, Zincati will reboot into this update when all\n\
-        interactive users have logged out, or in {}, whichever comes earlier.\n\
-        Please log out of all active sessions in order to let the auto-update process continue.",
+        If permitted by the update strategy, Zincati will reboot into this update when\n\
+        all interactive users have logged out, or in {}, whichever comes\n\
+        earlier. Please log out of all active sessions in order to let the auto-update\n\
+        process continue.",
         release_ver, time_till_reboot
     )
 }
@@ -517,7 +560,10 @@ mod tests {
 
     #[test]
     fn default_state() {
-        assert_eq!(UpdateAgentState::default(), UpdateAgentState::StartState);
+        assert_eq!(
+            UpdateAgentMachineState::default(),
+            UpdateAgentMachineState::StartState
+        );
     }
 
     #[test]
@@ -525,20 +571,20 @@ mod tests {
         let steady_interval = Duration::from_secs(DEFAULT_STEADY_INTERVAL_SECS);
         let default_interval = Duration::from_secs(DEFAULT_REFRESH_PERIOD_SECS);
 
-        let mut machine = UpdateAgentState::default();
-        assert_eq!(machine, UpdateAgentState::StartState);
+        let mut machine = UpdateAgentMachineState::default();
+        assert_eq!(machine, UpdateAgentMachineState::StartState);
 
         machine.initialized();
-        assert_eq!(machine, UpdateAgentState::Initialized);
+        assert_eq!(machine, UpdateAgentMachineState::Initialized);
 
         machine.reported_steady();
-        assert_eq!(machine, UpdateAgentState::ReportedSteady);
+        assert_eq!(machine, UpdateAgentMachineState::ReportedSteady);
 
         let state_change_time_before = LATEST_STATE_CHANGE.get();
         thread::sleep(time::Duration::from_secs(1));
         machine.no_new_update(); // ReportedSteady to NoNewUpdate.
         let state_change_time_after = LATEST_STATE_CHANGE.get();
-        assert_eq!(machine, UpdateAgentState::NoNewUpdate);
+        assert_eq!(machine, UpdateAgentMachineState::NoNewUpdate);
         assert_ne!(state_change_time_before, state_change_time_after);
         let (delay, should_jitter) = machine.get_refresh_delay(steady_interval);
         assert_eq!(delay, steady_interval);
@@ -548,7 +594,7 @@ mod tests {
         thread::sleep(time::Duration::from_secs(1));
         machine.no_new_update(); // NoNewUpdate to NoNewUpdate.
         let state_change_time_after = LATEST_STATE_CHANGE.get();
-        assert_eq!(machine, UpdateAgentState::NoNewUpdate);
+        assert_eq!(machine, UpdateAgentMachineState::NoNewUpdate);
         // Transitioning to own state not considered state change.
         assert_eq!(state_change_time_before, state_change_time_after);
 
@@ -560,14 +606,14 @@ mod tests {
         machine.update_available(update.clone());
         assert_eq!(
             machine,
-            UpdateAgentState::UpdateAvailable((update.clone(), 0))
+            UpdateAgentMachineState::UpdateAvailable((update.clone(), 0))
         );
 
         let (persistent_err, _) = machine.record_failed_deploy();
         assert_eq!(persistent_err, false);
         assert_eq!(
             machine,
-            UpdateAgentState::UpdateAvailable((update.clone(), 1))
+            UpdateAgentMachineState::UpdateAvailable((update.clone(), 1))
         );
         let (delay, should_jitter) = machine.get_refresh_delay(steady_interval);
         assert_eq!(delay, default_interval);
@@ -576,14 +622,17 @@ mod tests {
         machine.update_staged(update.clone());
         assert_eq!(
             machine,
-            UpdateAgentState::UpdateStaged((update.clone(), MAX_FINALIZE_POSTPONEMENTS))
+            UpdateAgentMachineState::UpdateStaged((update.clone(), MAX_FINALIZE_POSTPONEMENTS))
         );
 
         machine.update_finalized(update.clone());
-        assert_eq!(machine, UpdateAgentState::UpdateFinalized(update.clone()));
+        assert_eq!(
+            machine,
+            UpdateAgentMachineState::UpdateFinalized(update.clone())
+        );
 
         machine.end();
-        assert_eq!(machine, UpdateAgentState::EndState);
+        assert_eq!(machine, UpdateAgentMachineState::EndState);
     }
 
     #[test]
@@ -593,12 +642,12 @@ mod tests {
             checksum: "ostree-checksum".to_string(),
             age_index: None,
         };
-        let mut machine = UpdateAgentState::NoNewUpdate;
+        let mut machine = UpdateAgentMachineState::NoNewUpdate;
 
         machine.update_available(update.clone());
         assert_eq!(
             machine,
-            UpdateAgentState::UpdateAvailable((update.clone(), 0))
+            UpdateAgentMachineState::UpdateAvailable((update.clone(), 0))
         );
 
         // MAX-1 temporary failures.
@@ -607,14 +656,14 @@ mod tests {
             assert_eq!(persistent_err, false);
             assert_eq!(
                 machine,
-                UpdateAgentState::UpdateAvailable((update.clone(), attempt as u8))
+                UpdateAgentMachineState::UpdateAvailable((update.clone(), attempt as u8))
             )
         }
 
         // Persistent error threshold reached.
         let (persistent_err, _) = machine.record_failed_deploy();
         assert_eq!(persistent_err, true);
-        assert_eq!(machine, UpdateAgentState::NoNewUpdate);
+        assert_eq!(machine, UpdateAgentMachineState::NoNewUpdate);
     }
 
     #[test]
@@ -627,7 +676,7 @@ mod tests {
             checksum: "ostree-checksum".to_string(),
             age_index: None,
         };
-        let mut machine = UpdateAgentState::UpdateAvailable((update.clone(), 0));
+        let mut machine = UpdateAgentMachineState::UpdateAvailable((update.clone(), 0));
         let (delay, should_jitter) = machine.get_refresh_delay(steady_interval);
         assert_eq!(delay, default_interval);
         assert!(should_jitter);
@@ -635,7 +684,7 @@ mod tests {
         machine.update_staged(update.clone());
         assert_eq!(
             machine,
-            UpdateAgentState::UpdateStaged((update.clone(), MAX_FINALIZE_POSTPONEMENTS))
+            UpdateAgentMachineState::UpdateStaged((update.clone(), MAX_FINALIZE_POSTPONEMENTS))
         );
 
         // Set up empty interactive sessions.
@@ -644,7 +693,7 @@ mod tests {
         assert!(can_finalize);
         assert_eq!(
             machine,
-            UpdateAgentState::UpdateStaged((update.clone(), MAX_FINALIZE_POSTPONEMENTS))
+            UpdateAgentMachineState::UpdateStaged((update.clone(), MAX_FINALIZE_POSTPONEMENTS))
         );
         let (delay, should_jitter) = machine.get_refresh_delay(steady_interval);
         assert_eq!(delay, default_interval);
@@ -669,7 +718,7 @@ mod tests {
                 MAX_FINALIZE_POSTPONEMENTS.saturating_sub(finalization_attempt);
             assert_eq!(
                 machine,
-                UpdateAgentState::UpdateStaged((update.clone(), postponement_remaining))
+                UpdateAgentMachineState::UpdateStaged((update.clone(), postponement_remaining))
             );
             let (delay, should_jitter) = machine.get_refresh_delay(steady_interval);
             assert_eq!(delay, postponement_interval);
@@ -685,7 +734,10 @@ mod tests {
         // Reached 0 remaining postponements.
         let can_finalize = machine.handle_interactive_sessions(&interactive_sessions_present);
         assert!(can_finalize);
-        assert_eq!(machine, UpdateAgentState::UpdateStaged((update.clone(), 0)));
+        assert_eq!(
+            machine,
+            UpdateAgentMachineState::UpdateStaged((update.clone(), 0))
+        );
     }
 
     #[test]
